from enum import Enum
from importlib.metadata import version

from pydantic import BaseModel

from dash_pydantic_form.fields import fields
from dash_pydantic_form.form_layouts import AccordionFormLayout, FormLayout, StepsFormLayout, TabsFormLayout
from dash_pydantic_form.form_section import FormSection, Sections
from dash_pydantic_form.model_form import ModelForm
from dash_pydantic_utils import from_form_data, get_model_cls

BaseModel.__getitem__ = lambda self, key: self.__dict__.get(key)
BaseModel.to_plotly_json = lambda self: self.model_dump(mode="json")
Enum.to_plotly_json = lambda self: self.value

_css_dist = [
    {
        "relative_package_path": "pydf_styles.css",
        "namespace": "dash_pydantic_form",
    },
]
_js_dist = [
    {
        "relative_package_path": "pydf_clientside.js",
        "namespace": "dash_pydantic_form",
    },
    {
        "relative_package_path": "pydf_aggrid.js",
        "namespace": "dash_pydantic_form",
    },
]

__version__ = version(__package__)

__all__ = [
    "AccordionFormLayout",
    "FormLayout",
    "FormSection",
    "ModelForm",
    "Sections",
    "StepsFormLayout",
    "TabsFormLayout",
    "get_model_cls",
    "fields",
    "from_form_data",
    "__version__",
]



import logging

from .base_fields import CheckboxField as Checkbox
from .base_fields import ChecklistField as Checklist
from .base_fields import ChipField as Chip
from .base_fields import ChipGroupField as ChipGroup
from .base_fields import ColorField as Color
from .base_fields import DateField as Date
from .base_fields import DatetimeField as Datetime
from .base_fields import JsonField as Json
from .base_fields import MonthField as Month
from .base_fields import MultiSelectField as MultiSelect
from .base_fields import NumberField as Number
from .base_fields import PasswordField as Password
from .base_fields import RadioItemsField as RadioItems
from .base_fields import RangeField as Range
from .base_fields import RatingField as Rating
from .base_fields import SegmentedControlField as SegmentedControl
from .base_fields import SelectField as Select
from .base_fields import SliderField as Slider
from .base_fields import SwitchField as Switch
from .base_fields import TagsField as Tags
from .base_fields import TextareaField as Textarea
from .base_fields import TextField as Text
from .base_fields import TimeField as Time
from .base_fields import YearField as Year
from .dict_field import DictField as Dict
from .list_field import ListField as List
from .markdown_field import MarkdownField as Markdown
from .model_field import ModelField as Model
from .path_field import PathField as Path
from .quantity_field import QuantityField as Quantity
from .table_field import TableField as Table
from .transferlist_field import TransferListField as TransferList


def deprecated_field_factory(name: str, base_class: type):
    """Create a field class with a deprecation warning message."""
    old_post_init = getattr(base_class, "model_post_init", None)

    def post_init(self, _context):
        logging.warning(f"{name} is deprecated, use {base_class.__name__.removesuffix('Field')} instead.")
        if old_post_init is not None:
            old_post_init(self, _context)

    newclass = type(name, (base_class,), {"model_post_init": post_init})
    return newclass


# Backwards compatibility
ModelList = deprecated_field_factory("ModelList", List)
EditableTable = deprecated_field_factory("EditableTable", Table)

__all__ = [
    "Checkbox",
    "Checklist",
    "Chip",
    "ChipGroup",
    "Color",
    "Date",
    "Datetime",
    "Dict",
    "EditableTable",
    "Json",
    "List",
    "Markdown",
    "Model",
    "ModelList",
    "Month",
    "MultiSelect",
    "Number",
    "Password",
    "Path",
    "Quantity",
    "RadioItems",
    "Range",
    "Rating",
    "SegmentedControl",
    "Select",
    "Slider",
    "Switch",
    "Table",
    "Tags",
    "Textarea",
    "Text",
    "Time",
    "TransferList",
    "Year",
]



import base64
import io
import uuid
from datetime import date, datetime, time
from functools import partial
from typing import get_args

import dash_ag_grid as dag
import dash_mantine_components as dmc
from dash import MATCH, ClientsideFunction, Input, Output, State, callback, clientside_callback, dcc, html, no_update
from dash.development.base_component import Component
from pydantic import BaseModel, ConfigDict, Field
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import (
    BaseField,
    CheckboxField,
    ChecklistField,
    FieldsRepr,
    MonthField,
    MultiSelectField,
    RadioItemsField,
    SegmentedControlField,
    SelectField,
    TextareaField,
    TextField,
    YearField,
)
from dash_pydantic_form.fields.markdown_field import MarkdownField
from dash_pydantic_form.i18n import _
from dash_pydantic_form.ids import field_dependent_id
from dash_pydantic_utils import deep_merge, get_fullpath, get_non_null_annotation


class JSFunction(BaseModel):
    """JS function."""

    namespace: str
    function_name: str


class TableField(BaseField):
    """Editable table input field attributes and rendering."""

    fields_repr: FieldsRepr = Field(
        default=None,
        description="Fields representation, mapping between field name and field representation for the nested fields.",
    )
    with_upload: bool = Field(default=True, description="Whether to allow uploading a CSV file.")
    with_download: bool | None = Field(
        default=None,
        description="Whether to allow downloading the table as a CSV file."
        " If not set, it has the same value as `with_upload` by default.",
    )
    rows_editable: bool = Field(default=True, description="Whether to allow adding/removing rows.")
    table_height: int = Field(default=300, description="Table rows height in pixels.")
    column_defs_overrides: dict[str, dict] | None = Field(default=None, description="Ag-grid column_defs overrides.")
    dynamic_options: dict[str, JSFunction] | None = Field(
        default=None,
        description="Clientside function to use for dynamic options, defined per columnn."
        " The functions should take as argument the original options and the row data."
        " The functions should be defined on a sub-namespace of dash_clientside.",
    )
    grid_kwargs: dict = Field(
        default_factory=dict,
        description="Additional keyword arguments passed to the AGGrid instance. "
        "columnDefs passed here will not be considered, use column_defs_overrides.",
    )
    excluded_fields: list[str] | None = Field(default=None, description="Fields excluded from the sub-form")
    fields_order: list[str] | None = Field(default=None, description="Order of fields in the sub-form")

    full_width = True
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, _context):
        """Model post init."""
        super().model_post_init(_context)
        if self.fields_repr is None:
            self.fields_repr = {}
        if self.column_defs_overrides is None:
            self.column_defs_overrides = {}
        if self.with_download is None:
            self.with_download = self.with_upload
        if self.read_only:
            self.rows_editable = False
            self.with_upload = False
        if self.with_upload or self.with_download:
            try:
                import pandas  # noqa: F401
            except ModuleNotFoundError as exc:
                raise ValueError(
                    "The `with_upload` and `with_download` options are only available if pandas is installed."
                ) from exc

    class ids(BaseField.ids):
        """Model list field ids."""

        editable_table = partial(field_dependent_id, "_pydf-editable-table-table")
        upload_csv = partial(field_dependent_id, "_pydf-editable-table-upload")
        download_csv = partial(field_dependent_id, "_pydf-editable-table-download")
        download_csv_btn = partial(field_dependent_id, "_pydf-editable-table-download-btn")
        add_row = partial(field_dependent_id, "_pydf-editable-table-add-row")
        notification_wrapper = partial(field_dependent_id, "_pydf-editable-table-notification-wrapper")

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Create a form field of type Editable Table input to interact with the model."""
        value = self.get_value(item, field, parent) or []
        template = get_args(get_non_null_annotation(field_info.annotation))[0]
        if not issubclass(template, BaseModel):
            raise TypeError(f"Wrong type annotation for field {get_fullpath(parent, field)} to use Table.")

        required_fields = [f for f, f_info in template.model_fields.items() if f_info.is_required()]
        optional_fields = [f for f in template.model_fields if f not in required_fields]

        upload = []
        if self.with_upload:
            upload_ = dcc.Upload(
                id=self.ids.upload_csv(aio_id, form_id, field, parent=parent),
                children=dmc.Stack(
                    [
                        html.Div(
                            [
                                _("Drag & drop a csv file or "),
                                dmc.Anchor(_("select it"), href="#"),
                            ]
                        ),
                        dmc.Stack(
                            [
                                dmc.Text(_("CSV columns"), size="sm"),
                                dmc.Group(
                                    [
                                        dmc.Text(_("REQUIRED"), size="sm", style={"flexShrink": 0}),
                                        dmc.Group(
                                            [
                                                dmc.Badge(
                                                    f,
                                                    color="dark",
                                                    style={
                                                        "textTransform": "none",
                                                        "padding": "0 0.25rem",
                                                        "fontWeight": "normal",
                                                    },
                                                    radius="sm",
                                                )
                                                for f in required_fields
                                            ],
                                            gap="0.25rem",
                                        ),
                                    ],
                                    gap="0.5rem",
                                    wrap=False,
                                    align="start",
                                ),
                            ]
                            * bool(required_fields)
                            + [
                                dmc.Group(
                                    [
                                        dmc.Text(_("OPTIONAL"), size="sm", style={"flexShrink": 0}),
                                        dmc.Group(
                                            [
                                                dmc.Badge(
                                                    f,
                                                    color="dark",
                                                    style={
                                                        "textTransform": "none",
                                                        "padding": "0 0.25rem",
                                                        "fontWeight": "normal",
                                                    },
                                                    radius="sm",
                                                )
                                                for f in optional_fields
                                            ],
                                            gap="0.25rem",
                                        ),
                                    ],
                                    gap="0.5rem",
                                    wrap=False,
                                    align="start",
                                ),
                            ]
                            * bool(optional_fields),
                            gap="0.375rem",
                            mt="1rem",
                        ),
                    ],
                    gap=6,
                    align="start",
                    style={"margin": "0 auto", "width": "fit-content"},
                ),
                style={
                    "width": "100%",
                    "borderWidth": "1px",
                    "borderStyle": "dashed",
                    "borderRadius": "5px",
                    "padding": "1rem 1.5rem",
                },
            )
            upload = [
                dmc.Menu(
                    [
                        dmc.MenuTarget(dmc.Button(_("Upload CSV"), size="compact-sm")),
                        dmc.MenuDropdown(upload_),
                    ],
                    shadow="xl",
                    position="top-start",
                    styles={"dropdown": {"maxWidth": "min(90vw, 500px)"}},
                ),
            ]

        download = []
        if self.with_download:
            download = [
                dmc.Button(
                    _("Download CSV"),
                    id=self.ids.download_csv_btn(aio_id, form_id, field, parent=parent),
                    size="compact-sm",
                ),
                dcc.Download(id=self.ids.download_csv(aio_id, form_id, field, parent=parent)),
            ]

        add_row = []
        if self.rows_editable:
            add_row = [
                dmc.Button(
                    _("Add row"),
                    id=self.ids.add_row(aio_id, form_id, field, parent=parent),
                    size="compact-sm",
                ),
            ]

        def get_field_repr(field: str) -> BaseField | dict:
            from dash_pydantic_form.fields import get_default_repr

            if field in self.fields_repr:
                return self.fields_repr[field]
            return get_default_repr(template.model_fields[field])

        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)
        grid_kwargs = self.grid_kwargs
        grid_kwargs.pop("columnDefs", None)
        grid_kwargs.pop("rowData", None)
        column_defs = [  # Generate a column def depending on the field type
            self._generate_field_column(
                field_name=field_name,
                field_repr=get_field_repr(field_name),
                field_info=template.model_fields[field_name],
                required_field=field_name in required_fields,
                editable=not self.read_only,
            )
            for field_name in template.model_fields
            if field_name not in (self.excluded_fields or [])
        ]
        if self.fields_order:
            column_defs = [
                next(col for col in column_defs if col["field"] == field)
                for field in self.fields_order
                if field in template.model_fields
            ] + [col for col in column_defs if col["field"] not in self.fields_order]

        return html.Div(
            [
                html.Div(
                    (title is not None)
                    * [
                        dmc.Text(
                            [title]
                            + [
                                html.Span(
                                    " *", style={"color": "var(--input-asterisk-color, var(--mantine-color-error))"}
                                ),
                            ]
                            * self.is_required(field_info),
                            size="sm",
                            mt=3,
                            mb=5,
                            fw=500,
                            lh=1.55,
                        )
                    ]
                    + (title is not None and description is not None)
                    * [dmc.Text(description, size="xs", c="dimmed", mt=-5, mb=5, lh=1.2)],
                    style={"marginBottom": "-0.5rem" if title is not None else None},
                ),
                dag.AgGrid(
                    id=self.ids.editable_table(aio_id, form_id, field, parent=parent),
                    columnDefs=(
                        [
                            {
                                "headerName": "",
                                "cellRenderer": "PydfDeleteButton",
                                "lockPosition": "left",
                                "maxWidth": 35,
                                "filter": False,
                                "cellStyle": {
                                    "padding": 0,
                                    "display": "grid",
                                    "placeContent": "center",
                                },
                                "editable": False,
                            }
                        ]
                        # removes the delete button if add button is removed
                        if self.rows_editable
                        else []
                    )
                    + column_defs,
                    defaultColDef={"resizable": True, "sortable": True, "filter": True}
                    | grid_kwargs.pop("defaultColDef", {})
                    | {"editable": not self.read_only},
                    rowData=value,
                    columnSize=grid_kwargs.pop("columnSize", "responsiveSizeToFit"),
                    style=grid_kwargs.pop("style", {}) | {"height": self.table_height},
                    dashGridOptions={
                        "singleClickEdit": True,
                        "rowSelection": "multiple",
                        "stopEditingWhenCellsLoseFocus": True,
                    }
                    | grid_kwargs.pop("dashGridOptions", {})
                    | {
                        "suppressRowHoverHighlight": self.read_only,
                        "suppressRowClickSelection": self.read_only,
                    },
                    className=grid_kwargs.pop("className", "")
                    + " ag-theme-alpine ag-themed overflowing-ag-grid"
                    + (" read-only" if self.read_only else ""),
                    **grid_kwargs,
                ),
            ]
            + (
                [dmc.Group(add_row + upload + download)]
                if (self.rows_editable or self.with_upload or self.with_download)
                else []
            )
            + [
                dmc.JsonInput(
                    id=common_ids.value_field(aio_id, form_id, field, parent=parent),
                    value=value,
                    style={"display": "none"},
                ),
                html.Div(id=self.ids.notification_wrapper(aio_id, form_id, field, parent=parent)),
            ]
            * (not self.read_only),
            style={"display": "grid", "gap": "0.5rem", "gridTemplateColumns": "1fr"},
        )

    def _generate_field_column(  # noqa: PLR0913, PLR0912
        self,
        *,
        field_name: str,
        field_repr: BaseField | dict,
        field_info: FieldInfo,
        required_field: bool,
        editable: bool = True,
    ):
        """Takes a field and generates the 'columnDefs' dictionary for said field based on its type."""
        from dash_pydantic_form.fields import get_default_repr

        if isinstance(field_repr, dict):
            field_repr = get_default_repr(field_info, **field_repr)

        # Column_def no matter the type
        column_def = {
            "editable": editable,
            "field": field_name,
            "headerName": field_repr.get_title(field_info, field_name=field_name),
            "required": required_field,
            "cellClass": {
                "function": "(params.value == null || params.value === '') ? 'required_cell' : ''",
            }
            if required_field
            else None,
        }

        if description := field_repr.get_description(field_info):
            column_def.update({"headerTooltip": description})

        if field_info.default != PydanticUndefined:
            column_def["default_value"] = field_info.default
        if field_info.default_factory is not None:
            column_def["default_value"] = field_info.default_factory()

        # if select field, generate column of dropdowns
        if isinstance(
            field_repr, SelectField | SegmentedControlField | RadioItemsField | MultiSelectField | ChecklistField
        ):
            data = field_repr._additional_kwargs(field_info=field_info)["data"]
            options = [
                {
                    "value": x
                    if isinstance(x, str | int | float)
                    else (x["value"] if isinstance(x, dict) else x.value),
                    "label": x
                    if isinstance(x, str | int | float)
                    else (x["label"] if isinstance(x, dict) else x.label),
                }
                for x in data
            ]
            params = {k: v for k, v in field_repr.input_kwargs.items() if k not in ["data"]}
            if self.dynamic_options and field_name in self.dynamic_options:
                params["dynamicOptions"] = {
                    "namespace": self.dynamic_options[field_name].namespace,
                    "function_name": self.dynamic_options[field_name].function_name,
                }
            editor = "PydfMultiSelect" if isinstance(field_repr, MultiSelectField | ChecklistField) else "PydfDropdown"
            column_def.update(
                {
                    "cellEditor": {"function": editor},
                    "cellEditorPopup": False,
                    "cellEditorParams": {"options": options, **params},
                    "cellRenderer": "PydfOptionsRenderer",
                    "cellClass": {"function": "selectRequiredCell(params)"} if required_field else None,
                }
            )

        if isinstance(field_info, TextareaField | MarkdownField):
            column_def.update(
                {
                    "cellEditor": "agLargeTextCellEditor",
                    "cellEditorPopup": True,
                    "cellEditorParams": {"maxLength": 100, "rows": 10, "cols": 50},
                }
            )

        if isinstance(field_info, TextField | TextareaField | MarkdownField):
            column_def.update({"dtype": "str"})

        if isinstance(field_info, CheckboxField):
            column_def.update({"cellRenderer": "PydfCheckbox", "editable": False})

        annotation = get_non_null_annotation(field_info.annotation)
        if annotation in [int, float]:
            column_def.update({"filter": "agNumberColumnFilter"})

        if annotation in [date, datetime, time] or isinstance(field_repr, YearField | MonthField):
            picker_function = {
                date: "PydfDatePicker",
                datetime: "PydfDatetimePicker",
                time: "PydfTimePicker",
            }[annotation]

            if isinstance(field_repr, YearField):
                picker_function = "PydfYearPicker"
            elif isinstance(field_repr, MonthField):
                picker_function = "PydfMonthPicker"

            column_def.update(
                {
                    "cellEditor": {"function": picker_function},
                    "cellEditorPopup": annotation is datetime or isinstance(field_repr, YearField | MonthField),
                    "cellEditorParams": field_repr.input_kwargs,
                    "filter": "agDateColumnFilter",
                    "filterParams": {"comparator": {"function": "PydfDateComparator"}},
                },
            )

        # update with custom defs
        column_def = deep_merge(column_def, self.column_defs_overrides.get(field_name, {}))

        # default return base definition (text field)
        return column_def

    # Sync the JsonInput from the table data
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="syncTableJson"),
        Output(common_ids.value_field(MATCH, MATCH, MATCH, parent=MATCH), "value", allow_duplicate=True),
        Input(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "rowData"),
        Input(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "cellValueChanged"),
        prevent_initial_call=True,
    )

    @callback(
        Output(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "rowData", allow_duplicate=True),
        Output(ids.notification_wrapper(MATCH, MATCH, MATCH, parent=MATCH), "children", allow_duplicate=True),
        Input(ids.upload_csv(MATCH, MATCH, MATCH, parent=MATCH), "contents"),
        State(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "columnDefs"),
        prevent_initial_call=True,
    )
    def csv_to_table(contents, column_defs):
        """Output uploaded csv file to the editable table."""
        import pandas as pd

        if contents is not None:
            _, content_string = contents.split(",")

            decoded = base64.b64decode(content_string)
            data = pd.read_csv(
                io.StringIO(decoded.decode("utf-8")),
                dtype={f["field"]: f["dtype"] for f in column_defs if "field" in f and "dtype" in f},
            )
            required_columns = [col["field"] for col in column_defs if col.get("required")]
            if set(required_columns).issubset(data.columns):
                for col in column_defs:
                    if not (field := col.get("field")):
                        continue
                    if options := col.get("cellEditorParams", {}).get("options"):
                        values = [x["value"] for x in options]
                        options_dict = {x["label"]: x["value"] for x in options}
                        data[field] = data[field].where(data[field].isin(values), data[field].map(options_dict))

                return data.to_dict("records"), None

            return no_update, dmc.Notification(
                color="red",
                title=_("Wrong column names"),
                message=_("CSV upload failed, the file should contain the following columns: ")
                + f"{', '.join(required_columns)}",
                id=uuid.uuid4().hex,
                action="show",
            )
        return no_update, None

    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="tableAddRow"),
        Output(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "rowTransaction", allow_duplicate=True),
        Input(ids.add_row(MATCH, MATCH, MATCH, parent=MATCH), "n_clicks"),
        State(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "columnDefs"),
        prevent_initial_call=True,
    )

    @callback(
        Output(ids.download_csv(MATCH, MATCH, MATCH, parent=MATCH), "data"),
        Input(ids.download_csv_btn(MATCH, MATCH, MATCH, parent=MATCH), "n_clicks"),
        State(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "rowData"),
        prevent_initial_call=True,
    )
    def table_to_csv(n_clicks, table_data):
        """Send the table data to the user as a CSV file."""
        if n_clicks and table_data:
            import pandas as pd

            data_df = pd.DataFrame(table_data)
            return dcc.send_data_frame(data_df.to_csv, "table_data.csv", index=False, encoding="utf-8")
        return no_update

    clientside_callback(
        """ data => !(Array.isArray(data) && data.length) """,
        Output(ids.download_csv_btn(MATCH, MATCH, MATCH, parent=MATCH), "disabled"),
        Input(ids.editable_table(MATCH, MATCH, MATCH, parent=MATCH), "rowData"),
    )



import itertools
from copy import deepcopy
from functools import partial
from typing import Literal

import dash_mantine_components as dmc
from dash.development.base_component import Component
from dash_iconify import DashIconify

from dash_pydantic_form.form_section import FormSection
from dash_pydantic_form.ids import form_base_id
from dash_pydantic_utils import deep_merge

from .form_layout import FormLayout, Position


class TabsFormLayout(FormLayout):
    """Tabs form layout.

    Parameters
    ----------
    render_kwargs: dict | None
        Kwargs to pass to the render method.
    sections: list[FormSection]
        List of form sections.
    remaining_fields_position: Position
        Position of the remaining fields, one of 'top', 'bottom' or 'none'.
    """

    class ids:
        """Tabs form ids."""

        tabs = partial(form_base_id, "_pydf-tabs")

    sections: list[FormSection]
    remaining_fields_position: Position = "top"
    layout: Literal["tabs"] = "tabs"

    def render(  # noqa: PLR0913
        self,
        *,
        field_inputs: dict[str, Component],
        aio_id: str,
        form_id: str,
        path: str,
        **_kwargs,
    ):
        """Render the sections in an accordion."""
        kwargs = deepcopy(self.render_kwargs or {})
        value = self.sections[0].name
        for section in self.sections:
            if section.default_open:
                value = section.name
                break

        tabs_styles = deep_merge({"panel": {"padding": "1rem 0.5rem"}}, kwargs.pop("styles", {}))

        tabs = dmc.Tabs(
            [
                dmc.TabsList(
                    [
                        dmc.TabsTab(
                            dmc.Text(
                                ([DashIconify(icon=section.icon)] if section.icon else []) + [section.name],
                                style={"display": "flex", "alignItems": "center", "gap": "0.5rem"},
                            ),
                            value=section.name,
                        )
                        for section in self.sections
                    ]
                ),
                *[
                    dmc.TabsPanel(
                        self.grid([field_inputs[field] for field in section.fields if field in field_inputs]),
                        value=section.name,
                    )
                    for section in self.sections
                ],
            ],
            value=value,
            styles=tabs_styles,
            id=self.ids.tabs(aio_id, form_id, path),
            **kwargs,
        )

        fields_not_in_sections = set(field_inputs) - set(itertools.chain.from_iterable(s.fields for s in self.sections))

        if self.remaining_fields_position == "none" or not fields_not_in_sections:
            return [tabs]
        if self.remaining_fields_position == "top":
            return [self.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm"), tabs]
        return [tabs, self.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm")]



from .accordion_form_layout import AccordionFormLayout
from .form_layout import FormLayout
from .steps_form_layout import StepsFormLayout
from .tabs_form_layout import TabsFormLayout

__all__ = ["AccordionFormLayout", "FormLayout", "StepsFormLayout", "TabsFormLayout"]



import dataclasses as dc
import uuid
import warnings
from copy import deepcopy
from functools import partial
from types import UnionType
from typing import Annotated, Literal, Union, get_args, get_origin, overload

import dash_mantine_components as dmc
from dash import (
    ALL,
    MATCH,
    ClientsideFunction,
    Input,
    Output,
    State,
    callback,
    clientside_callback,
    ctx,
    dcc,
    html,
    no_update,
)
from dash.development.base_component import Component, rd
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from dash_pydantic_utils import (
    SEP,
    Type,
    get_fullpath,
    get_model_cls,
    get_subitem,
    get_subitem_cls,
    handle_discriminated,
    is_subclass,
    model_construct_recursive,
)

from . import ids as common_ids
from .fields import BaseField, fields
from .form_layouts import FormLayout
from .i18n import _, language_context
from .ids import form_base_id

Children_ = Component | str | int | float
Children = Children_ | list[Children_]
SectionRender = Literal["accordion", "tabs", "steps"]
Position = Literal["top", "bottom", "none"]


class ModelFormIdsFactory:
    """Factory functions for model form ids."""

    form = partial(form_base_id, "_pydf-form")
    main = partial(form_base_id, "_pydf-main")
    restore_wrapper = partial(form_base_id, "_pydf-restore-wrapper")
    restore_btn = partial(form_base_id, "_pydf-restore-btn")
    cancel_restore_btn = partial(form_base_id, "_pydf-cancel-restore-btn")
    wrapper = partial(common_ids.field_dependent_id, "_pydf-wrapper")
    errors = partial(form_base_id, "_pydf-errors")
    model_store = partial(form_base_id, "_pydf-model-store")
    form_specs_store = partial(form_base_id, "_pydf-form-specs-store")


@dc.dataclass(frozen=True)
class ModelFormIds:
    """Model form ids."""

    form: dict[str, str]
    main: dict[str, str]
    restore_wrapper: dict[str, str]
    restore_btn: dict[str, str]
    cancel_restore_btn: dict[str, str]
    errors: dict[str, str]
    model_store: dict[str, str]
    form_specs_store: dict[str, str]

    @classmethod
    def from_basic_ids(cls, aio_id: str, form_id: str) -> "ModelFormIds":
        """Instanciation from aio_id and form_id."""
        return cls(*(getattr(ModelFormIdsFactory, id_field.name)(aio_id, form_id) for id_field in dc.fields(cls)))


class IdAccessor:
    """Descriptor for handling access to either instances or the factory of model form ids via ``ModelForm`` class."""

    @overload
    def __get__(self, obj: "ModelForm", objtype=None) -> ModelFormIds: ...
    @overload
    def __get__(self, obj: None, objtype=None) -> type[ModelFormIdsFactory]: ...
    def __get__(self, obj, objtype=None):
        """Returns the ``ModelFormIdsFactory`` class if accessed via the ``ModelForm`` class directly (ModelForm.ids)
        or an instance of ``ModelFormIds`` if accessed via an instance of ``ModelForm`` (ModelForm(my_model).ids).
        """
        if obj is None:
            # access via class
            return ModelFormIdsFactory

        if isinstance(obj, ModelForm):
            # access via instance
            return obj._ids

        raise RuntimeError("IdAccessor should only be defined on ModelForm or an instance of ModelForm")

    def __set__(self, obj: "ModelForm", value: ModelFormIds | tuple[str, str]):
        """Sets another set of model form ids to a ``ModelForm`` object."""
        if isinstance(value, tuple):
            value = ModelFormIds.from_basic_ids(*value)

        obj._ids = value


class ModelForm(html.Div):
    """Create a Dash form from a pydantic model.

    Parameters
    ----------
    item: BaseModel | type[BaseModel]
        The model to create the form from, can be the model class or an instance of the class.
        If the class is passed, the form will be empty. If an instance is passed, the form will be pre-filled
        with existing values.
    aio_id: str | None
        All-in-one component ID. A pseudo-random string will be auto-generated if not provided.
    form_id: str | None
        Form ID, can be used to create multiple forms on the same page. When working with databases
        this could be the document / record ID. A pseudo-random string will be auto-generated if not provided.
    form_cols: int
        Number of columns in the form, defaults to 4.
    fields_repr: dict[str, dict | BaseField] | None
        Mapping between field name and field representation. If not provided, default field
        representations will be used based on the field annotation.
        See `fields.get_default_repr`.
    sections: FormLayout | None
        DEPRECATED: Use form_layout.
    form_layout: FormLayout | None
        Defines how to render the form layout, built-in layouts are `accordion`, `tabs` and `steps`.
        If not provided, default to rendering the fields in a grid.
    submit_on_enter: bool
        Whether to submit the form on enter. Default False.
        Note: this may break the behaviour of some fields (e.g. editable table), use with caution.
    excluded_fields: list[str] | None
        List of field names to exclude from the form altogether, optional.
    container_kwargs: dict | None
        Additional kwargs to pass to the containing div.
    read_only: bool | None
        Whether the form should be read only.
        True/False set all the fields to read only or not. None keeps the field setting.
    debounce: int | None
        Debounce the form output in milliseconds.
    debounce_inputs: int | None
        DEPRECATED: Use debounce.
    locale: str | None
        Locale to render the buttons and helpers in, currently English and French are supported.
        If left to None, will default to system locale, and fallback to English.
    cols: int
        Deprecated, use `form_cols` instead.
    data_model: type[BaseModel] | Annotated[UnionType, FieldInfo] | None
        The data model or union of models to create the form from, this is mostly used for discriminated unions.
    fields_order: list[str] | None
        List of field names to order the fields in the form. The fields will be displayed in the order provided.
        All fields not in the list will be displayed in ther model order, after the ones defined here.
    store_progress: bool | Literal["local", "session"]
        Whether to store the progress of the form in the local store, to allow picking up where the user left off.
        If set to True or "local" will store to local storage, if set to "session" will store to session storage.
    """

    ids = IdAccessor()
    _ids: ModelFormIds

    def __init__(  # noqa: PLR0912, PLR0913, PLR0915
        self,
        item: BaseModel | type[BaseModel] | Annotated[UnionType, FieldInfo],
        aio_id: str | None = None,
        form_id: str | None = None,
        path: str = "",
        form_cols: int = 4,
        fields_repr: dict[str, dict | BaseField] | None = None,
        sections: FormLayout | None = None,
        form_layout: FormLayout | None = None,
        submit_on_enter: bool = False,
        discriminator: str | None = None,
        excluded_fields: list[str] | None = None,
        container_kwargs: dict | None = None,
        read_only: bool | None = None,
        debounce: int | None = None,
        debounce_inputs: int | None = None,
        locale: str = None,
        cols: int = None,
        data_model: type[BaseModel] | Annotated[UnionType, FieldInfo] | None = None,
        fields_order: list[str] | None = None,
        store_progress: bool | Literal["local", "session"] = False,
        restore_behavior: Literal["auto", "notify"] = "notify",
    ) -> None:
        if data_model is None:
            data_model = type(item) if isinstance(item, BaseModel) else item
        if is_subclass(item, BaseModel):
            item = item.model_construct()
        if not isinstance(item, BaseModel):
            item = None
        if get_origin(data_model) is Annotated:
            if discriminator is None:
                discriminator = next((f.discriminator for f in get_args(data_model) if isinstance(f, FieldInfo)), None)
            data_model = get_args(data_model)[0]

        aio_id = aio_id or str(uuid.UUID(int=rd.randint(0, 2**128)))
        form_id = form_id or str(uuid.UUID(int=rd.randint(0, 2**128)))
        self.ids = ModelFormIds.from_basic_ids(aio_id, form_id)

        if cols is not None:
            warnings.warn("cols is deprecated, use form_cols instead", DeprecationWarning, stacklevel=1)
            form_cols = cols
        if sections is not None:
            warnings.warn("sections is deprecated, use form_layout instead", DeprecationWarning, stacklevel=1)
            if form_layout is None:
                form_layout = sections
        if debounce_inputs is not None:
            warnings.warn("debounce_inputs is deprecated, use debounce instead", DeprecationWarning, stacklevel=1)
            if debounce is None:
                debounce = debounce_inputs

        fields_repr = fields_repr or {}

        subitem_cls, disc_vals = self.get_discriminated_subitem_cls(
            item=item, path=path, discriminator=discriminator, data_model=data_model
        )
        with language_context(locale):
            field_inputs = self.render_fields(
                item=item,
                aio_id=aio_id,
                form_id=form_id,
                path=path,
                subitem_cls=subitem_cls,
                disc_vals=disc_vals,
                fields_repr=fields_repr,
                excluded_fields=excluded_fields,
                discriminator=discriminator,
                read_only=read_only,
                form_cols=form_cols,
            )
            # Re-order fields as per fields_order
            if fields_order:
                field_inputs = {f: field_inputs[f] for f in fields_order if f in field_inputs} | {
                    f: r for f, r in field_inputs.items() if f not in fields_order
                }

            if not form_layout:
                children = [FormLayout.grid(list(field_inputs.values()))]
            else:
                children = form_layout.render(
                    field_inputs=field_inputs,
                    aio_id=aio_id,
                    form_id=form_id,
                    path=path,
                    read_only=read_only,
                    form_cols=form_cols,
                )

            children.extend(
                self._get_meta_children(
                    fields_repr=fields_repr,
                    form_layout=form_layout,
                    aio_id=aio_id,
                    form_id=form_id,
                    path=path,
                    form_cols=form_cols,
                    data_model=data_model,
                    restore_behavior=restore_behavior,
                )
            )

        container_kwargs = container_kwargs or {}
        style = {"--pydf-form-cols": f"{form_cols}"} | container_kwargs.pop("style", {}) | {"position": "relative"}
        if not path:
            style |= {"containerType": "inline-size"}

        super().__init__(
            children=html.Div(children, id=ModelFormIdsFactory.wrapper(aio_id, form_id, discriminator or "", path)),
            style=style,
            **(
                {
                    "id": ModelFormIdsFactory.form(aio_id, form_id, path),
                    "data-submitonenter": submit_on_enter,
                    "data-storeprogress": store_progress,
                    "data-getvalues": None,
                    "data-restored": None,
                    "data-update": None,
                    "data-submit": None,
                    "data-debounce": debounce,
                }
                if not path
                else {}
            ),
            **container_kwargs,
        )

    @staticmethod
    def get_discriminated_subitem_cls(
        *,
        item: BaseModel | None,
        path: str,
        discriminator: str | None,
        data_model: type[BaseModel] | UnionType,
    ) -> tuple[type[BaseModel], tuple]:
        """Get the subitem of a model at a given parent, handling type unions."""
        subitem_cls = (
            get_subitem_cls(item.__class__, path, item=item) if is_subclass(data_model, BaseModel) else data_model
        )

        # Handle type unions
        disc_vals = None
        discriminator_value = None
        if Type.classify(subitem_cls, discriminator) == Type.DISCRIMINATED_MODEL:
            subitem = get_subitem(item, path) if item is not None else None
            discriminator_value = None if subitem is None else getattr(subitem, discriminator, None)
            subitem_cls, disc_vals = handle_discriminated(
                item.__class__, path, subitem_cls, discriminator, discriminator_value
            )

        return subitem_cls, disc_vals

    @staticmethod
    def render_fields(  # noqa: PLR0913
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        path: str,
        subitem_cls: type[BaseModel],
        disc_vals: list[str],
        fields_repr: dict[str, dict | BaseField],
        excluded_fields: list[str],
        discriminator: str | None,
        read_only: bool | None,
        form_cols: int,
    ) -> dict[str, Component]:
        """Render each field in the form."""
        from dash_pydantic_form.fields import get_default_repr

        excluded_fields = (excluded_fields or []) + subitem_cls.model_config.get("private_fields", [])

        field_inputs = {}
        for field_name, field_info in subitem_cls.model_fields.items():
            if field_name in (excluded_fields or []):
                continue
            more_kwargs = {"form_cols": form_cols}
            if read_only:
                more_kwargs["read_only"] = read_only
            # If discriminating field, ensure all discriminator values are shown
            # Also add required metadata for discriminator callback
            if disc_vals and field_name == discriminator:
                field_info = deepcopy(field_info)  # noqa: PLW2901
                field_info.annotation = Literal[disc_vals]
                more_kwargs |= {"n_cols": "var(--pydf-form-cols)", "field_id_meta": "discriminator"}
            if field_name in fields_repr:
                if isinstance(fields_repr[field_name], dict):
                    field_repr = get_default_repr(field_info, **(fields_repr[field_name] | more_kwargs))
                else:
                    field_repr = fields_repr[field_name]
                    if more_kwargs:
                        field_repr = field_repr.__class__(**(field_repr.model_dump() | more_kwargs))
            else:
                field_repr = get_default_repr(field_info, **more_kwargs)

            field_inputs[field_name] = field_repr.render(
                item=item,
                aio_id=aio_id,
                form_id=form_id,
                field=field_name,
                parent=path,
                field_info=field_info,
            )

        return field_inputs

    @classmethod
    def _get_meta_children(  # noqa: PLR0913
        cls,
        *,
        fields_repr: dict[str, dict | BaseField],
        form_layout: FormLayout | None,
        aio_id: str,
        form_id: str,
        path: str,
        form_cols: int,
        data_model: type[BaseModel] | UnionType,
        restore_behavior: Literal["auto", "notify"],
    ):
        """Get 'meta' form children used for passing data to callbacks."""
        children = []
        if not path:
            children.append(
                html.Div(
                    cls.get_restore_data_overlay(aio_id, form_id),
                    id=cls.ids.restore_wrapper(aio_id, form_id),
                    style={"display": "none"},
                    **{"data-behavior": restore_behavior},
                )
            )
            children.append(dcc.Store(id=cls.ids.main(aio_id, form_id)))
            children.append(dcc.Store(id=cls.ids.errors(aio_id, form_id)))
            if is_subclass(data_model, BaseModel):
                model_name = str(data_model)
            elif get_origin(data_model) in [Union, UnionType]:
                model_name = [str(x) for x in get_args(data_model)]
            else:
                raise ValueError("data_model must be a pydantic BaseModel or Union of models")
            children.append(dcc.Store(data=model_name, id=cls.ids.model_store(aio_id, form_id)))

        fields_repr_dicts = (
            {
                field_name: field_repr if isinstance(field_repr, dict) else field_repr.model_dump(mode="json")
                for field_name, field_repr in fields_repr.items()
            }
            if fields_repr
            else None
        )

        children.append(
            dcc.Store(
                data={
                    "form_layout": form_layout.model_dump(mode="json") if form_layout else None,
                    "fields_repr": fields_repr_dicts,
                    "form_cols": form_cols,
                },
                id=cls.ids.form_specs_store(aio_id, form_id, path),
            )
        )

        return children

    @classmethod
    def get_restore_data_overlay(cls, aio_id: str, form_id: str):
        """Get the overlay for restoring data."""
        return dmc.Stack(
            dmc.Card(
                [
                    dmc.Text(_("Found previous form data. Do you want to restore it?"), size="sm", mb="1.25rem"),
                    dmc.Group(
                        [
                            dmc.Button(
                                _("Restore"),
                                size="compact-sm",
                                id=cls.ids.restore_btn(aio_id, form_id),
                            ),
                            dmc.Button(
                                _("Cancel"),
                                size="compact-sm",
                                variant="outline",
                                id=cls.ids.cancel_restore_btn(aio_id, form_id),
                            ),
                        ],
                        justify="center",
                    ),
                ],
                shadow="lg",
                withBorder=True,
            ),
            pos="absolute",
            style={"inset": 0, "backdropFilter": "blur(3px)", "zIndex": 1},
            bg="color-mix(in srgb, var(--mantine-color-body), #0000 25%)",
            p="2rem",
            align="center",
        )


clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="getValues"),
    Output(ModelForm.ids.main(MATCH, MATCH), "data"),
    Input(common_ids.value_field(MATCH, MATCH, ALL, ALL, ALL), "value"),
    Input(common_ids.checked_field(MATCH, MATCH, ALL, ALL, ALL), "checked"),
    Input(fields.Dict.ids.item_key(MATCH, MATCH, ALL, ALL, ALL), "value"),
    Input(BaseField.ids.visibility_wrapper(MATCH, MATCH, ALL, ALL, ALL), "style"),
    Input(ModelForm.ids.form(MATCH, MATCH), "data-getvalues"),
    State(ModelForm.ids.form(MATCH, MATCH), "id"),
    State(ModelForm.ids.form(MATCH, MATCH), "data-storeprogress"),
    State(ModelForm.ids.main(MATCH, MATCH), "data"),
    State(ModelForm.ids.restore_wrapper(MATCH, MATCH), "id"),
    State(ModelForm.ids.restore_wrapper(MATCH, MATCH), "data-behavior"),
    State(ModelForm.ids.form(MATCH, MATCH), "data-debounce"),
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="restoreData"),
    Output(ModelForm.ids.form(MATCH, MATCH), "data-update", allow_duplicate=True),
    Output(ModelForm.ids.restore_wrapper(MATCH, MATCH), "style", allow_duplicate=True),
    Output(ModelForm.ids.form(MATCH, MATCH), "data-restored", allow_duplicate=True),
    Input(ModelForm.ids.restore_btn(MATCH, MATCH), "n_clicks"),
    State(ModelForm.ids.form(MATCH, MATCH), "data-restored"),
    prevent_initial_call=True,
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="cancelRestoreData"),
    Output(ModelForm.ids.restore_wrapper(MATCH, MATCH), "style", allow_duplicate=True),
    Output(ModelForm.ids.form(MATCH, MATCH), "data-restored", allow_duplicate=True),
    Output(ModelForm.ids.form(MATCH, MATCH), "data-getvalues", allow_duplicate=True),
    Input(ModelForm.ids.cancel_restore_btn(MATCH, MATCH), "n_clicks"),
    prevent_initial_call=True,
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="listenToSubmit"),
    Output(ModelForm.ids.form(MATCH, MATCH), "id"),
    Input(ModelForm.ids.form(MATCH, MATCH), "id"),
    State(ModelForm.ids.form(MATCH, MATCH), "data-submitonenter"),
    prevent_initial_call=True,
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="displayErrors"),
    Output(common_ids.value_field(MATCH, MATCH, ALL, ALL, ALL), "error"),
    Input(ModelForm.ids.errors(MATCH, MATCH), "data"),
    State(common_ids.value_field(MATCH, MATCH, ALL, ALL, ALL), "id"),
    prevent_initial_call=True,
)


@callback(
    Output(ModelForm.ids.wrapper(MATCH, MATCH, MATCH, MATCH), "children", allow_duplicate=True),
    Input(ModelForm.ids.form(MATCH, MATCH, MATCH), "data-update"),
    State(ModelForm.ids.model_store(MATCH, MATCH, MATCH), "data"),
    State(ModelForm.ids.form_specs_store(MATCH, MATCH, MATCH), "data"),
    prevent_initial_call=True,
)
def update_data(form_data: dict, model_name: str | list[str], form_specs: dict):
    """Update contents with ids.form data-update."""
    if not form_data:
        return no_update
    return update_form_wrapper_contents(form_data, None, model_name, form_specs)


@callback(
    Output(ModelForm.ids.wrapper(MATCH, MATCH, MATCH, MATCH), "children", allow_duplicate=True),
    Input(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "discriminator"), "value"),
    State(ModelForm.ids.main(MATCH, MATCH), "data"),
    State(ModelForm.ids.model_store(MATCH, MATCH), "data"),
    State(ModelForm.ids.form_specs_store(MATCH, MATCH, MATCH), "data"),
    prevent_initial_call=True,
)
def update_discriminated(val, form_data: dict, model_name: str | list[str], form_specs: dict):
    """Update contents when discriminator input changes."""
    path: str = get_fullpath(ctx.triggered_id["parent"], ctx.triggered_id["field"])
    discriminator = ctx.triggered_id["field"]
    parts = path.split(SEP)
    # Update the form data with the new value as it wouldn't have been updated yet
    pointer = form_data
    for i, part in enumerate(parts):
        if i == len(parts) - 1:
            pointer[part] = val
        if part.isdigit():
            pointer = list(pointer.values())[int(part)] if isinstance(pointer, dict) else pointer[int(part)]
        else:
            pointer = pointer[part]
    return update_form_wrapper_contents(form_data, discriminator, model_name, form_specs)


def update_form_wrapper_contents(
    form_data: dict,
    discriminator: str | None,
    model_name: str | list[str],
    form_specs: dict,
):
    """Update the form wrapper contents."""
    # Create an instance of the model with the form data using model_construct_recursive
    # to build it out as much as possible without failing on validation
    if isinstance(model_name, str):
        model_cls = get_model_cls(model_name)
    else:
        if not (disc_val := form_data.get(discriminator)):
            return no_update
        model_union = [get_model_cls(x) for x in model_name]
        model_cls = next(
            (x for x in model_union if x.model_fields[discriminator].default == disc_val),
            None,
        )
        if model_cls is None:
            return no_update

    item = model_construct_recursive(form_data, model_cls)

    # Retrieve fields-repr and form_layout from the stored data
    fields_repr: dict[str, BaseField] = recursive_load_fields_repr(form_specs["fields_repr"] or {})
    form_layout = FormLayout.load(**form_specs["form_layout"]) if form_specs["form_layout"] else None

    form = ModelForm(
        item=item,
        aio_id=ctx.triggered_id["aio_id"],
        form_id=ctx.triggered_id["form_id"],
        path=ctx.triggered_id["parent"],
        discriminator=discriminator,
        form_layout=form_layout,
        fields_repr=fields_repr,
        form_cols=form_specs["form_cols"],
        data_model=None if isinstance(model_name, str) else Union[tuple(model_union)],  # noqa: UP007
    )

    return form.children.children


def recursive_load_fields_repr(fields_repr: dict):
    """Recursively load fields repr."""
    final = deepcopy(fields_repr)
    for k, v in fields_repr.items():
        if "fields_repr" in v:
            v["fields_repr"] = recursive_load_fields_repr(v["fields_repr"])
        try:
            final[k] = BaseField.load(v)
        except KeyError:
            final[k] = v
    return final



import inspect
import json
import logging
import os
from collections.abc import Callable
from enum import Enum, EnumMeta
from functools import partial
from textwrap import TextWrapper
from types import UnionType
from typing import Annotated, Any, ClassVar, Literal, Union, get_args, get_origin

import dash_mantine_components as dmc
from dash import ALL, MATCH, ClientsideFunction, Input, Output, State, clientside_callback, html
from dash.development.base_component import Component
from pydantic import AfterValidator, BaseModel, ConfigDict, Field, model_serializer
from pydantic.fields import FieldInfo
from pydantic.types import annotated_types
from pydantic_core import PydanticUndefined

from dash_pydantic_form import ids as common_ids
from dash_pydantic_utils import (
    SEP,
    Type,
    get_all_subclasses,
    get_fullpath,
    get_model_value,
    get_non_null_annotation,
)

CHECKED_COMPONENTS = [
    dmc.Checkbox,
    dmc.Switch,
    dmc.Chip,
]
CHECKED_CHILDREN_COMPONENTS = [
    dmc.Chip,
]
NO_LABEL_COMPONENTS = [
    dmc.SegmentedControl,
    dmc.ChipGroup,
    dmc.RangeSlider,
    dmc.Slider,
    dmc.ColorPicker,
    dmc.Rating,
]
MAX_OPTIONS_INLINE = 4

FilterOperator = Literal["==", "!=", "in", "not in", "array_contains", "array_contains_any"]
VisibilityFilter = tuple[str, FilterOperator, Any]


class BaseField(BaseModel):
    """Base field representation class."""

    base_component: ClassVar[type[Component] | None] = None
    reserved_attributes: ClassVar = ("value", "label", "description", "id", "required")
    full_width: ClassVar[bool] = False

    title: str | Component | None = Field(
        default=None, description="Field label, overrides the title defined in the pydantic Field."
    )
    description: str | Component | None = Field(
        default=None, description="Field helper text, overrides the description defined in the pydantic Field."
    )
    required: bool | None = Field(
        default=None,
        description="Whether to display a required asterisk. If not provided, uses pydantic's field `is_required`.",
    )
    n_cols: int | float | str | None = Field(
        default=None,
        description="Number of form columns the fields spans. "
        "If an int is provided, the field will span that many columns. "
        "If a float is provided, it should be between 0 and 1 and represent a share of the available columns. "
        "If a string is provided, it should represent the css to go with `grid-column: span <x>`. "
        "If None, will default to half the available columns.",
    )
    visible: bool | VisibilityFilter | list[VisibilityFilter] | None = Field(
        default=None,
        description=(
            "Define visibility conditions based on other form fields.\n"
            "Accepts a boolean, a 3-tuple or list of 3-tuples with format: (field, operator, value).\n"
            "The available operators are:\n"
            "* '=='\n"
            "* '!='\n"
            "* 'in'\n"
            "* 'not in'\n"
            "* 'array_contains'\n"
            "* 'array_contains_any'\n"
            "NOTE: The field in the 3-tuples is a ':' separated path relative to the current field's "
            "level of nesting.\n"
            "If you need to reference a field from a parent or the root "
            "use the special values `_parent_` or `_root_`.\n"
            "E.g. visible=('_root_:first_name', '==', 'Bob')"
            ""
        ),
    )
    input_kwargs: dict | None = Field(
        default=None,
        description=(
            "Arguments to be passed to the underlying rendered component. "
            "NOTE: these are updated with extra arguments passed to the field."
        ),
    )
    field_id_meta: str | None = Field(default=None, description="Optional str to be set in the field id's 'meta' key.")
    read_only: bool = Field(default=False, description="Read only field.")

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    @classmethod
    def __pydantic_init_subclass__(cls):
        """Add docstring in subclasses."""
        tw = TextWrapper(width=89, initial_indent="    ", subsequent_indent="    ")
        result = (cls.__doc__ or "") + "\n\nParameters\n----------\n"
        for field_name, field_info in cls.model_fields.items():
            annotation = str(field_info.annotation)
            description = tw.fill(field_info.description) if field_info.description else "    (missing description)"
            result += f"{field_name}: {annotation}\n{description}\n"

        cls.__doc__ = result

    def model_post_init(self, _context):
        """Model post init."""
        if self.n_cols is None:
            self.n_cols = "var(--pydf-form-cols)" if self.full_width else "calc(var(--pydf-form-cols) / 2)"
        if isinstance(self.n_cols, float) and (self.n_cols < 0 or self.n_cols > 1):
            raise ValueError("Field n_cols must be between 0 and 1 when using a float.")
        if self.input_kwargs is None:
            self.input_kwargs = {}
        if self.model_extra:
            self.input_kwargs.update(self.model_extra)
        if self.base_component:
            valid_input_kwargs = {
                k: v for k, v in self.input_kwargs.items() if k in inspect.signature(self.base_component).parameters
            }
            ignored_kwargs = set(self.input_kwargs) - set(valid_input_kwargs)
            self.input_kwargs = valid_input_kwargs
            if ignored_kwargs:
                logging.debug("The following kwargs were ignored for %s: %s", self.__class__.__name__, ignored_kwargs)
        if self.read_only:
            self.input_kwargs["className"] = self.input_kwargs.get("className", "") + " read-only"
        if self.field_id_meta is None:
            self.field_id_meta = ""

    @property
    def n_cols_css(self):
        """Get number of columns CSS variable."""
        if isinstance(self.n_cols, str):
            return self.n_cols
        if isinstance(self.n_cols, float):
            return f"calc(var(--pydf-form-cols) * {self.n_cols})"
        return f"{self.n_cols}"

    class ids:
        """Form ids."""

        visibility_wrapper = partial(common_ids.field_dependent_id, "_pydf-field-visibility-wrapper")

    @model_serializer(mode="wrap")
    def serialize_with_cls_name(self, base_serializer, info):
        """Add class name in model_dump."""
        return {"__class__": str(self.__class__)} | base_serializer(self, info)

    @classmethod
    def load(cls, data):
        """Create a field from a dictionary."""
        data = data.copy()
        str_repr = data.pop("__class__")
        field_cls = next(c for c in get_all_subclasses(BaseField) if str(c) == str_repr)
        return field_cls(**data)

    def render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo,
    ) -> Component:
        """Render the field."""
        """Create a form input to interact with the field, and conditional visibility wrapper."""
        title = None
        if os.getenv("DEBUG"):
            title = f"Field path: {get_fullpath(parent, field)}"

        inputs = self._render(
            item=item, aio_id=aio_id, form_id=form_id, field=field, parent=parent, field_info=field_info
        )
        visible = self.visible

        if visible is None or visible is True:
            return html.Div(
                inputs, className="pydantic-form-field", style={"--pydf-field-cols": self.n_cols_css}, title=title
            )

        if field_info.default == PydanticUndefined and field_info.default_factory is None:
            logging.warning(
                "Conditional visibility is set on a field without default value, "
                f"this will likely lead to validation errors. Field: {get_fullpath(parent, field)}"
            )

        if visible is False:
            return html.Div(inputs, style={"display": "none"}, title=title)

        if isinstance(visible, tuple) and isinstance(visible[0], str):
            visible = [visible]

        for i, vis in enumerate(visible):
            inputs, title = self._add_visibility_wrapper(
                inputs=inputs,
                aio_id=aio_id,
                form_id=form_id,
                item=item,
                visibility=vis,
                parent=parent,
                field=field,
                index=i,
                n_visibility_fields=len(visible),
                title=title,
            )

        return inputs

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Create a form input to interact with the field."""
        if not self.base_component:
            raise NotImplementedError("This is an abstract class.")

        value = self.get_value(item, field, parent)

        if (
            self.read_only
            and self.base_component is not None
            and (
                "readOnly" not in inspect.signature(self.base_component).parameters
                # NOTE: readOnly not working on SegmentedControl in 0.14.5
                or self.base_component is dmc.SegmentedControl
            )
        ):
            return self._render_read_only(value, field, field_info)

        id_ = (common_ids.checked_field if self.base_component in CHECKED_COMPONENTS else common_ids.value_field)(
            aio_id, form_id, field, parent, meta=self.field_id_meta
        )
        value_kwarg = (
            {
                "children" if self.base_component in CHECKED_CHILDREN_COMPONENTS else "label": self.get_title(
                    field_info, field_name=field
                )
            }
            | ({"checked": value} if value is not None else {})
            if self.base_component in CHECKED_COMPONENTS
            else (
                {
                    "label": self.get_title(field_info, field_name=field),
                    "description": self.get_description(field_info),
                    "required": self.is_required(field_info),
                    "readOnly": self.read_only,
                }
                if self.base_component not in NO_LABEL_COMPONENTS
                else {}
            )
            | ({"value": value} if value is not None else {})
        )

        component = self.base_component(
            id=id_,
            **self.input_kwargs
            | self._additional_kwargs(item=item, aio_id=aio_id, field=field, parent=parent, field_info=field_info)
            | value_kwarg,
        )

        if self.base_component not in NO_LABEL_COMPONENTS:
            return component

        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)
        return dmc.Stack(
            (title is not None)
            * [
                dmc.Text(
                    [title]
                    + [
                        html.Span(" *", style={"color": "var(--input-asterisk-color, var(--mantine-color-error))"}),
                    ]
                    * self.is_required(field_info),
                    size="sm",
                    mt=3,
                    mb=5,
                    fw=500,
                    lh=1.55,
                )
            ]
            + (title is not None and description is not None)
            * [dmc.Text(description, size="xs", c="dimmed", mt=-5, mb=5, lh=1.2)]
            + [component],
            gap=0,
        )

    def _render_read_only(self, value: Any, field: str, field_info: FieldInfo):
        """Render a read only field."""
        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)

        outputs = dmc.Stack(
            (title is not None) * [dmc.Text(title, size="sm", mt=3, mb=5, fw=500, lh=1.55)]
            + (title is not None and description is not None)
            * [dmc.Text(description, size="xs", c="dimmed", mt=-5, mb=5, lh=1.2)],
            gap=0,
        )

        value_repr = self._get_value_repr(value, field_info)

        outputs.children.append(
            dmc.Paper(
                value_repr,
                withBorder=True,
                radius="sm",
                p="0.375rem 0.75rem",
                style={
                    "display": "flex",
                    "alignItems": "center",
                    "gap": "0.5rem",
                    "background": "var(--mantine-color-default)",
                    "borderColor": "color-mix(in srgb, var(--mantine-color-default-border), transparent 80%)",
                },
            )
        )

        return outputs

    @staticmethod
    def _get_value_repr(value: Any, field_info: FieldInfo):
        val_type = Type.classify(field_info.annotation)
        value_repr = str(value)
        if val_type == Type.SCALAR:
            if isinstance(value, bool):
                value_repr = dmc.Checkbox(checked=value, disabled=True, className="read-only")
            if isinstance(value, Enum):
                value_repr = value.name
            if value is None:
                value_repr = "-"
        elif val_type in [Type.SCALAR_LIST, Type.LITERAL_LIST]:
            value_repr = [dmc.Badge(str(val), variant="light", radius="sm", tt="unset") for val in value]

        return value_repr

    @staticmethod
    def _get_dependent_field_and_parent(dependent_field: str, parent: str):
        """Get the dependent field and parent.

        Manages the special pointers _root_ and _parent_.
        """
        dependent_parent_parts = parent.split(SEP) if parent else []
        for part in dependent_field.split(SEP)[:-1]:
            if part == "_root_":
                dependent_parent_parts = []
            elif part == "_parent_":
                dependent_parent_parts = dependent_parent_parts[:-1]
            else:
                dependent_parent_parts.append(part)

        dependent_parent = SEP.join(dependent_parent_parts) if dependent_parent_parts else ""
        dependent_field = dependent_field.split(SEP)[-1]

        return dependent_parent, dependent_field

    def _add_visibility_wrapper(  # noqa: PLR0913
        self,
        *,
        inputs,
        aio_id: str,
        form_id: str,
        item: BaseModel,
        visibility: tuple,
        parent: str,
        field: str,
        index: int,
        n_visibility_fields: int,
        title: str,
    ):
        """Wrap the inputs with a layer of togglable visibility."""
        dependent_field, operator, expected_value = visibility
        dependent_parent, dependent_field = self._get_dependent_field_and_parent(dependent_field, parent)

        current_value = self.get_value(item, dependent_field, dependent_parent)
        if isinstance(current_value, Enum):
            current_value = current_value.value
        if os.getenv("DEBUG"):
            keyword = "Visible" if index == 0 else "   AND"
            title += f"\n{keyword}: {get_fullpath(dependent_parent , dependent_field)}" f" {operator} {expected_value}"

        inputs = html.Div(
            inputs,
            id=self.ids.visibility_wrapper(
                aio_id,
                form_id,
                dependent_field,
                parent=dependent_parent,
                meta=f"{get_fullpath(parent, field)}|{operator}|{json.dumps(expected_value)}",
            ),
            className="pydantic-form-field",
            style={
                "display": None if self.check_visibility(current_value, operator, expected_value) else "none",
                "--pydf-field-cols": self.n_cols_css,
            },
            title=title if index == n_visibility_fields - 1 else None,
        )

        return inputs, title

    def get_title(self, field_info: FieldInfo, field_name: str | None = None) -> str:
        """Get the input title."""
        if self.title is not None:
            return self.title or None
        if field_info.title is not None:
            return field_info.title
        return field_name.replace("_", " ").title() if isinstance(field_name, str) else None

    def get_description(self, field_info: FieldInfo) -> str:
        """Get the input description."""
        return self.description or field_info.description

    def is_required(self, field_info: FieldInfo) -> bool:
        """Get the required status of the field."""
        return (self.required or field_info.is_required()) and not self.read_only

    @classmethod
    def _additional_kwargs(cls, **_kwargs) -> dict:
        """Additional kwargs."""
        return {}

    @staticmethod
    def check_visibility(value: Any, operator: str, expected_value: Any) -> bool:
        """Check whether a field should be visible based on value, operator and expected value."""
        if operator == "==":
            return value == expected_value
        if operator == "!=":
            return value != expected_value
        if operator == "in":
            return value in expected_value
        if operator == "not in":
            return value not in expected_value
        if operator == "array_contains":
            return expected_value in value
        if operator == "array_contains_any":
            return bool(set(value).intersection(expected_value))
        raise ValueError(f"Invalid operator: {operator}")

    @classmethod
    def get_value(cls, item: BaseModel, field: str, parent: str) -> Any:
        """Get the value of a model (parent, field) pair. Defined to allow overriding."""
        return get_model_value(item, field, parent)


class TextField(BaseField):
    """Text field."""

    base_component = dmc.TextInput


class TextareaField(BaseField):
    """Textarea field."""

    base_component = dmc.Textarea


class NumberField(BaseField):
    """Number field."""

    base_component = dmc.NumberInput

    def _additional_kwargs(self, field_info: FieldInfo, **_kwargs) -> dict:
        kwargs = {}
        for meta in field_info.metadata:
            if isinstance(meta, annotated_types.Ge):
                kwargs["min"] = meta.ge
            if isinstance(meta, annotated_types.Gt):
                kwargs["min"] = meta.gt + 1e-12
            if isinstance(meta, annotated_types.Le):
                kwargs["max"] = meta.le
            if isinstance(meta, annotated_types.Lt):
                kwargs["max"] = meta.lt - 1e-12

        return kwargs


class RatingField(BaseField):
    """Rating field."""

    base_component = dmc.Rating


class PasswordField(BaseField):
    """Password field."""

    base_component = dmc.PasswordInput


class JsonField(BaseField):
    """Json field."""

    base_component = dmc.JsonInput


class ColorField(BaseField):
    """Color field."""

    base_component = dmc.ColorInput

    # @staticmethod
    # def _get_value_repr(value: Any, field_info: FieldInfo):  # noqa: ARG004
    #     return dmc.Group(
    #         [dmc.Badge(color=value, p=0, h="1rem", w="1rem", radius="xs"), dmc.Text(value, size="sm")],
    #         gap="sm",
    #         align="center",
    #     )


class SliderField(BaseField):
    """Slider field."""

    base_component = dmc.Slider


class RangeField(BaseField):
    """Range field."""

    base_component = dmc.RangeSlider


class CheckboxField(BaseField):
    """Checkbox field."""

    base_component = dmc.Checkbox

    @classmethod
    def get_value(cls, item: BaseModel, field: str, parent: str) -> Any:
        """Default value to False if None."""
        value = super().get_value(item, field, parent)
        if value is None:
            value = False
        return value


class SwitchField(CheckboxField):
    """Switch field."""

    base_component = dmc.Switch


class ChipField(CheckboxField):
    """Switch field."""

    base_component = dmc.Chip


class DateField(BaseField):
    """Date field."""

    base_component = dmc.DateInput

    def model_post_init(self, _context):
        """Add defaults for date input."""
        super().model_post_init(_context)
        self.input_kwargs.setdefault("valueFormat", "YYYY-MM-DD")


class TimeField(BaseField):
    """Time field."""

    base_component = dmc.TimeInput


class DatetimeField(BaseField):
    """Datetime field."""

    base_component = dmc.DateTimePicker

    def model_post_init(self, _context):
        """Add defaults for date input."""
        super().model_post_init(_context)
        self.input_kwargs.setdefault("valueFormat", "YYYY-MM-DD HH:mm")


class YearField(BaseField):
    """Year field."""

    base_component = dmc.YearPickerInput


class MonthField(BaseField):
    """Year-month field."""

    base_component = dmc.MonthPickerInput

    def model_post_init(self, _context):
        """Add defaults for date input."""
        super().model_post_init(_context)
        self.input_kwargs.setdefault("valueFormat", "YYYY-MM")


class SelectField(BaseField):
    """Select field."""

    data_getter: str | None = Field(
        default=None, description="Function to retrieve a list of options. This function takes no argument."
    )
    options_labels: dict | None = Field(
        default=None, description="Mapper from option to label. Especially useful for Literals and Enums."
    )

    base_component = dmc.Select

    getters: ClassVar[dict[str, Callable]] = {}

    @classmethod
    def register_data_getter(cls, data_getter: Callable[[], list[str]], name: str | None = None):
        """Register a data_getter."""
        name = name or str(data_getter)
        if name in cls.getters:
            logging.warning("Data getter %s already registered for Select field.", name)
        cls.getters[name] = data_getter

    def _get_data(self, field_info: FieldInfo, **kwargs) -> list[dict]:
        """Gets option list from annotations."""
        non_null_annotation = get_non_null_annotation(field_info.annotation)
        data = self._get_data_list(non_null_annotation=non_null_annotation, **kwargs)
        options = self._format_data(data, **kwargs)

        values, filtered = [], []
        for option in options:
            if option["value"] not in values:
                values.append(option["value"])
                filtered.append(option)

        return filtered

    def _get_data_list(
        self,
        non_null_annotation: type,
        **kwargs,
    ) -> list[dict]:
        """Get list of possible values from annotation."""
        data = self._get_data_list_recursive(non_null_annotation, **kwargs)
        return data

    def _get_data_list_recursive(self, non_null_annotation: type, **kwargs) -> list:
        """Get list of possible values from annotation recursively."""
        data = []
        # if the annotation is a union of types, recursively calls this function on each type.
        if get_origin(non_null_annotation) in [Union, UnionType]:
            data.extend(
                sum(
                    [self._get_data_list_recursive(sub_annotation) for sub_annotation in get_args(non_null_annotation)],
                    [],
                )
            )

        elif get_origin(non_null_annotation) is list:
            annotation_args = get_args(non_null_annotation)
            if len(annotation_args) == 1:
                return self._get_data_list_recursive(annotation_args[0], **kwargs)
        elif get_origin(non_null_annotation) == Literal:
            data = list(get_args(non_null_annotation))
        elif isinstance(non_null_annotation, EnumMeta):
            data = [{"value": x.value, "label": x.name} for x in non_null_annotation]

        return data

    def _format_data(self, data, **_kwargs):
        """Formats the list of options into a `value, label` pair."""
        if self.options_labels:
            return [
                {"value": x["value"], "label": self.options_labels.get(x["value"], x["label"])}
                if isinstance(x, dict)
                else {"value": x, "label": str(self.options_labels.get(x, x))}
                for x in data
            ]

        return [x if isinstance(x, dict) else {"value": x, "label": x} for x in data]

    @property
    def data_gotten(self):
        """Get data from data_getter."""
        if self.data_getter:
            if self.data_getter not in self.getters:
                raise ValueError(f"Data getter {self.data_getter} not registered, use `register_data_getter`.")
            return self.getters[self.data_getter]()
        return None

    def _additional_kwargs(self, **kwargs) -> dict:
        """Retrieve data from Literal annotation if data is not present in input_kwargs."""
        return {"data": self.data_gotten or self.input_kwargs.get("data", self._get_data(**kwargs))}

    def _get_value_repr(self, value: Any, field_info: FieldInfo):
        value_repr = super()._get_value_repr(value, field_info)
        data = self._get_data(field_info)

        def _get_label(value, data, value_repr):
            if isinstance(value, Enum):
                value = value.value
            option = next(
                (x for x in data if (x.get("value") if isinstance(x, dict) else getattr(x, "value", None)) == value),
                None,
            )
            label = (
                option.get("label")
                if isinstance(option, dict)
                else getattr(option, "label", getattr(option, "children", None))
            )
            return label if label is not None else value_repr

        if Type.classify(field_info.annotation) == Type.SCALAR:
            return _get_label(value, data, value_repr)
        if Type.classify(field_info.annotation) == Type.SCALAR_LIST:
            return [dmc.Badge(_get_label(x, data, x), radius="sm", variant="light", tt="unset") for x in value]
        return value_repr


class MultiSelectField(SelectField):
    """MultiSelect field."""

    base_component = dmc.MultiSelect

    @classmethod
    def get_value(cls, item: BaseModel, field: str, parent: str) -> Any:
        """Handle the fact dmc.MultiSelect only allows string values."""
        value = super().get_value(item, field, parent)
        if isinstance(value, list):
            value = [str(x) for x in value]

        return value


class TagsField(SelectField):
    """Tags field."""

    base_component = dmc.TagsInput


class RadioItemsField(SelectField):
    """Radio items field."""

    base_component = dmc.RadioGroup
    orientation: Literal["horizontal", "vertical"] | None = Field(
        default=None,
        description="Orientation of the chip group, defaults to None which will adapt based on the number of items.",
    )

    def _additional_kwargs(self, *, field: str = None, field_info: FieldInfo, **kwargs) -> dict:
        """Retrieve data from Literal annotation if data is not present in input_kwargs."""
        kwargs = super()._additional_kwargs(field_info=field_info, **kwargs)
        data = kwargs["data"] or []
        children = [
            x
            if isinstance(x, dmc.Radio)
            else (dmc.Radio(**x) if isinstance(x, dict) else dmc.Radio(label=str(x), value=x))
            for x in data
        ]
        mt = "5px" if self.get_title(field_info, field_name=field) and self.get_description(field_info) else 0
        if self.orientation == "horizontal" or (self.orientation is None and len(data) <= MAX_OPTIONS_INLINE):
            return {"children": dmc.Group(children, mt=mt, py="0.5rem")}
        return {"children": dmc.Stack(children, mt=mt, py="0.25rem")}


class ChecklistField(MultiSelectField):
    """Checklist field."""

    base_component = dmc.CheckboxGroup
    orientation: Literal["horizontal", "vertical"] | None = Field(
        default=None,
        description="Orientation of the chip group, defaults to None which will adapt based on the number of items.",
    )

    def _additional_kwargs(self, *, field: str = None, field_info: FieldInfo, **kwargs) -> dict:
        """Retrieve data from Literal annotation if data is not present in input_kwargs."""
        kwargs = super()._additional_kwargs(field_info=field_info, **kwargs)
        data = kwargs["data"] or []
        children = [
            x
            if isinstance(x, dmc.Checkbox)
            else (dmc.Checkbox(**x) if isinstance(x, dict) else dmc.Checkbox(label=str(x), value=x))
            for x in data
        ]
        mt = "5px" if self.get_title(field_info, field_name=field) and self.get_description(field_info) else 0
        if self.orientation == "horizontal" or (self.orientation is None and len(data) <= MAX_OPTIONS_INLINE):
            return {"children": dmc.Group(children, mt=mt, py="0.5rem")}
        return {"children": dmc.Stack(children, mt=mt, py="0.25rem")}


class SegmentedControlField(SelectField):
    """Segmented control field."""

    base_component = dmc.SegmentedControl

    def model_post_init(self, _context):
        """Add default style."""
        super().model_post_init(_context)
        self.input_kwargs.setdefault("style", {"alignSelf": "baseline"})


class ChipGroupField(SelectField):
    """Segmented control field."""

    base_component = dmc.ChipGroup
    orientation: Literal["horizontal", "vertical"] | None = Field(
        default=None,
        description="Orientation of the chip group, defaults to None which will adapt based on the number of items.",
    )

    def _additional_kwargs(self, *, field: str = None, field_info: FieldInfo, **kwargs) -> dict:
        """Retrieve data from Literal annotation if data is not present in input_kwargs."""
        kwargs = super()._additional_kwargs(field_info=field_info, **kwargs)
        data = kwargs["data"] or []
        children = [
            x
            if isinstance(x, dmc.Chip)
            else (dmc.Chip(x["label"], value=x["value"]) if isinstance(x, dict) else dmc.Chip(str(x), value=str(x)))
            for x in data
        ]
        mt = "5px" if self.get_title(field_info, field_name=field) and self.get_description(field_info) else 0
        if self.orientation == "horizontal" or (self.orientation is None and len(data) <= MAX_OPTIONS_INLINE):
            return {"children": dmc.Group(children, mt=mt, py="0.5rem", gap="0.5rem")}
        return {"children": dmc.Stack(children, mt=mt, py="0.25rem", gap="0.5rem")}


def check_fields_repr(fields_repr):
    """Fields repr validator function."""
    if fields_repr is None:
        return None
    if any(not isinstance(x, dict | BaseField) for x in fields_repr.values()):
        raise ValueError("fields_repr must be a dict of BaseField or dict")
    return fields_repr


# NOTE: The following is defined to allow using custom serializer on nested fields.
# Pydantic doesn't work well when using dict[str, dict | BaseField] for fields_repr
# as it then tries to serialize the nested fields with BaseField rather than their own class.
# See https://github.com/pydantic/pydantic/issues/11563
FieldsRepr = Annotated[dict[str, Any] | None, AfterValidator(check_fields_repr)]


clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="updateFieldVisibility"),
    Output(BaseField.ids.visibility_wrapper(MATCH, MATCH, MATCH, MATCH, ALL), "style"),
    Input(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, ALL), "value"),
    Input(common_ids.checked_field(MATCH, MATCH, MATCH, MATCH, ALL), "checked"),
    State(BaseField.ids.visibility_wrapper(MATCH, MATCH, MATCH, MATCH, ALL), "style"),
    prevent_initial_call=True,
)



from functools import partial
from typing import Literal, get_args

import dash_mantine_components as dmc
from dash import dcc, html
from dash.development.base_component import Component
from dash_iconify import DashIconify
from plotly.io.json import to_json_plotly
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import BaseField
from dash_pydantic_form.fields.list_field import ListField
from dash_pydantic_form.form_layouts.form_layout import FormLayout
from dash_pydantic_form.i18n import _
from dash_pydantic_utils import Type, get_fullpath


class DictField(ListField):
    """Dict field for dict of models or scalars."""

    render_type: Literal["accordion", "modal", "scalar"] = "accordion"

    class ids(ListField.ids):
        """Dict field ids."""

        item_key = partial(common_ids.field_dependent_id, "_pydf-dict-field-item-key")

    @classmethod
    def accordion_item(  # noqa: PLR0913
        cls,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        fields_repr: dict[str, dict | BaseField] | None = None,
        form_layout: FormLayout | None = None,
        items_deletable: bool = True,
        read_only: bool | None = None,
        value: str | None = None,
        **kwargs,
    ):
        """Create an item with bare forms for the model dict field."""
        contents = super().accordion_item(
            item=item,
            aio_id=aio_id,
            form_id=form_id,
            field=field,
            parent=parent,
            index=index,
            value=kwargs.pop("value", None),
            fields_repr=fields_repr,
            form_layout=form_layout,
            items_deletable=items_deletable,
            read_only=read_only,
            **kwargs,
        )
        contents.children[0].children[0] = cls.key_input(
            aio_id, form_id, field, parent, index, key=value, read_only=read_only, w="calc(100% - 3.5rem)"
        )
        return contents

    @classmethod
    def modal_item(  # noqa: PLR0913
        cls,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        value: str | None = None,
        opened: bool = False,
        fields_repr: dict[str, dict | BaseField] | None = None,
        form_layout: FormLayout | None = None,
        items_deletable: bool = True,
        read_only: bool | None = None,
        **kwargs,
    ):
        """Create an item with bare forms for the model dict field."""
        contents = super().modal_item(
            item=item,
            aio_id=aio_id,
            form_id=form_id,
            field=field,
            parent=parent,
            index=index,
            value=kwargs.pop("value", None),
            opened=opened,
            fields_repr=fields_repr,
            form_layout=form_layout,
            items_deletable=items_deletable,
            read_only=read_only,
            **kwargs,
        )
        contents.children.children[0] = cls.key_input(
            aio_id, form_id, field, parent, index, key=value, read_only=read_only, style={"flex": 1}
        )
        return contents

    @classmethod
    def scalar_item(  # noqa: PLR0913
        cls,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        value: str | None = None,
        items_deletable: bool = True,
        read_only: bool | None = None,
        input_kwargs: dict,
        **kwargs,
    ):
        """Create an item for a scalar dict."""
        contents = super().scalar_item(
            item=item,
            aio_id=aio_id,
            form_id=form_id,
            field=field,
            parent=parent,
            index=index,
            items_deletable=items_deletable,
            read_only=read_only,
            input_kwargs=input_kwargs,
            value="",
            **kwargs,
        )
        contents.children = [
            cls.key_input(
                aio_id, form_id, field, parent, index, key=value, read_only=read_only, style={"flex": "0 0 30%"}
            ),
            dmc.Text(":", pl="0.25rem", pr="0.5rem", fw=500, lh=2),
        ] + contents.children
        return contents

    @classmethod
    def key_input(  # noqa: PLR0913
        cls,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str,
        index: int,
        key: str | None = None,
        read_only: bool | None = None,
        **input_kwargs,
    ):
        """Create an input for the key of a dict item."""
        return dmc.TextInput(
            placeholder="Key",
            id=cls.ids.item_key(aio_id, form_id, field, parent=parent, meta=index),
            leftSection=DashIconify(icon="fe:key"),
            value=key,
            readOnly=read_only,
            **input_kwargs,
        )

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo,
    ) -> Component:
        """Create a form field of type checklist to interact with the model field."""
        type_ = Type.classify(field_info.annotation, field_info.discriminator)
        if type_ in [Type.MODEL_LIST, Type.MODEL_DICT] and self.render_type == "scalar":
            raise ValueError("Cannot render model list as scalar")
        if (
            type_
            not in [Type.MODEL_LIST, Type.MODEL_DICT, Type.DISCRIMINATED_MODEL_LIST, Type.DISCRIMINATED_MODEL_DICT]
            and self.render_type != "scalar"
        ):
            raise ValueError("Cannot render non model list as non scalar")

        discriminator = (
            get_args(get_args(field_info.annotation)[1])[1].discriminator
            if type_ in [Type.DISCRIMINATED_MODEL_LIST, Type.DISCRIMINATED_MODEL_DICT]
            else None
        )

        value: list = self.get_value(item, field, parent) or []

        wrapper_class_name = "pydf-model-list-wrapper" + (" required" if self.is_required(field_info) else "")
        contents = self.render_type_items_mapper(self.render_type)(
            item=item,
            aio_id=aio_id,
            form_id=form_id,
            field=field,
            parent=parent,
            value=value,
            fields_repr=self.fields_repr,
            form_layout=self.form_layout,
            items_deletable=self.items_deletable,
            read_only=self.read_only,
            input_kwargs=self.input_kwargs,
            discriminator=discriminator,
            form_cols=self.form_cols,
            wrapper_class_name=wrapper_class_name,
            wrapper_kwargs=self.wrapper_kwargs,
        )

        # Create a template item to be used clientside when adding new items
        template = self.render_type_item_mapper(self.render_type)(
            item=item.__class__.model_construct(),
            aio_id=aio_id,
            form_id=form_id,
            field=field,
            parent=parent,
            index="{{" + get_fullpath(parent, field).replace(":", "|") + "}}",
            value="-",
            opened=True,
            fields_repr=self.fields_repr,
            form_layout=self.form_layout,
            items_deletable=self.items_deletable,
            read_only=self.read_only,
            input_kwargs=self.input_kwargs,
            discriminator=discriminator,
            form_cols=self.form_cols,
        )
        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)

        return dmc.Stack(
            bool(title)
            * [
                dmc.Stack(
                    bool(title)
                    * [
                        dmc.Text(
                            [title]
                            + [
                                html.Span(
                                    " *", style={"color": "var(--input-asterisk-color, var(--mantine-color-error))"}
                                ),
                            ]
                            * self.is_required(field_info),
                            size="sm",
                            mt=3,
                            fw=500,
                            lh=1.55,
                        )
                    ]
                    + (bool(title) and bool(description)) * [dmc.Text(description, size="xs", c="dimmed", lh=1.2)],
                    gap=0,
                )
            ]
            + [
                contents,
                dcc.Store(
                    data=to_json_plotly(template), id=self.ids.template_store(aio_id, form_id, field, parent=parent)
                ),
            ]
            + self.items_creatable
            * [
                html.Div(
                    [
                        dmc.Button(
                            _("Add"),
                            leftSection=DashIconify(icon="carbon:add", height=16),
                            size="compact-sm",
                            id=self.ids.add(aio_id, form_id, field, parent=parent),
                        ),
                    ],
                ),
            ],
            className="pydantic-form-field",
            style={"--pydf-field-cols": "var(--pydf-form-cols)"},
            gap="0.5rem",
        )



from functools import partial
from typing import Literal

import dash_mantine_components as dmc
from dash import (
    MATCH,
    ClientsideFunction,
    Input,
    Output,
    clientside_callback,
)
from dash.development.base_component import Component
from dash_iconify import DashIconify
from pydantic import BaseModel, Field, field_validator
from pydantic.fields import FieldInfo

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import BaseField, FieldsRepr
from dash_pydantic_form.form_layouts.form_layout import FormLayout
from dash_pydantic_form.i18n import _
from dash_pydantic_utils import get_fullpath


class ModelField(BaseField):
    """Model field, used for nested BaseModel.

    Optional attributes:
    * render_type (one of 'accordion', 'modal', default 'accordion')
    * fields_repr, mapping between field name and field representation
    * form_layout, FormLayout for the NestedModel form
    * form_cols, number of columns in the form
    * excluded_fields, list of field names to exclude from the form altogether
    * fields_order, list of field names in the order they should be rendered
    """

    render_type: Literal["accordion", "modal", "simple"] = Field(
        default="accordion", description="How to render the model field, one of 'accordion', 'modal', 'simple'."
    )
    fields_repr: FieldsRepr = Field(
        default=None,
        description="Fields representation, mapping between field name and field representation for the nested fields.",
    )
    form_layout: FormLayout | None = Field(default=None, description="Sub-form layout.")
    form_cols: int = Field(default=4, description="Number of columns in the form.")
    excluded_fields: list[str] | None = Field(default=None, description="Fields excluded from the sub-form")
    fields_order: list[str] | None = Field(default=None, description="Order of fields in the sub-form")

    full_width = True

    @field_validator("form_layout", mode="before")
    @classmethod
    def validate_form_layout(cls, v):
        """Validate form layout."""
        if isinstance(v, FormLayout) or v is None:
            return v
        if isinstance(v, dict):
            return FormLayout.load(**v)
        raise ValueError("form_layout must be a FormLayout or a dict that can be converted to a FormLayout")

    def model_post_init(self, _context):
        """Model post init."""
        super().model_post_init(_context)
        if self.fields_repr is None:
            self.fields_repr = {}
        if self.form_layout is None and self.model_extra and self.model_extra.get("sections") is not None:
            self.form_layout = self.model_extra["sections"]

    class ids(BaseField.ids):
        """Model field ids."""

        edit = partial(common_ids.field_dependent_id, "_pydf-model-field-edit")
        modal = partial(common_ids.field_dependent_id, "_pydf-model-field-modal")
        modal_save = partial(common_ids.field_dependent_id, "_pydf-model-field-modal-save")

    def simple_item(  # noqa: PLR0913
        self,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Model field simple render."""
        from dash_pydantic_form import ModelForm

        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)
        return dmc.Box(
            [
                dmc.Stack(
                    (title is not None) * [dmc.Text(title, size="sm", mt=3, mb=5, fw=500, lh=1.55)]
                    + (title is not None and description is not None)
                    * [dmc.Text(description, size="xs", c="dimmed", mt=-5, mb=5, lh=1.2)],
                    gap=0,
                ),
                ModelForm(
                    item=item,
                    aio_id=aio_id,
                    form_id=form_id,
                    path=get_fullpath(parent, field),
                    fields_repr=self.fields_repr,
                    form_layout=self.form_layout,
                    read_only=self.read_only,
                    form_cols=self.form_cols,
                    excluded_fields=self.excluded_fields,
                    fields_order=self.fields_order,
                ),
            ]
        )

    def modal_item(  # noqa: PLR0913
        self,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Model field modal render."""
        from dash_pydantic_form import ModelForm

        title = self.get_title(field_info, field_name=field)
        new_parent = get_fullpath(parent, field)
        return dmc.Paper(
            dmc.Group(
                [
                    dmc.Text(
                        title,
                        style={
                            "flex": 1,
                            "overflow": "hidden",
                            "textOverflow": "ellipsis",
                            "whiteSpace": "nowrap",
                        },
                    ),
                    dmc.Group(
                        [
                            dmc.ActionIcon(
                                DashIconify(icon="carbon:edit", height=16),
                                variant="light",
                                size="sm",
                                id=self.ids.edit(aio_id, form_id, field, parent=parent),
                            ),
                        ],
                        gap="0.25rem",
                    ),
                    dmc.Modal(
                        [
                            ModelForm(
                                item=item,
                                aio_id=aio_id,
                                form_id=form_id,
                                path=new_parent,
                                fields_repr=self.fields_repr,
                                form_layout=self.form_layout,
                                read_only=self.read_only,
                                form_cols=self.form_cols,
                                excluded_fields=self.excluded_fields,
                                fields_order=self.fields_order,
                            ),
                            dmc.Group(
                                dmc.Button(
                                    _("Save"),
                                    leftSection=DashIconify(icon="carbon:save"),
                                    id=self.ids.modal_save(aio_id, form_id, field, parent=parent),
                                ),
                                justify="right",
                                mt="sm",
                            ),
                        ],
                        title=title,
                        id=self.ids.modal(aio_id, form_id, field, parent=parent),
                        style={"--modal-size": "min(calc(100vw - 4rem), 1150px)"},
                        styles={"content": {"containerType": "inline-size"}},
                    ),
                ],
                gap="sm",
                align="top",
            ),
            withBorder=True,
            radius="sm",
            p="xs",
            className="pydf-model-list-modal-item",
        )

    def accordion_item(  # noqa: PLR0913
        self,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Model field accordion item render."""
        from dash_pydantic_form import ModelForm

        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)
        return dmc.Accordion(
            dmc.AccordionItem(
                value="item",
                children=[
                    dmc.AccordionControl(dmc.Text(title)),
                    dmc.AccordionPanel(
                        [
                            *([dmc.Text(description, size="xs", c="dimmed")] * bool(title) * bool(description)),
                            ModelForm(
                                item=item,
                                aio_id=aio_id,
                                form_id=form_id,
                                path=get_fullpath(parent, field),
                                fields_repr=self.fields_repr,
                                form_layout=self.form_layout,
                                discriminator=field_info.discriminator,
                                read_only=self.read_only,
                                form_cols=self.form_cols,
                                excluded_fields=self.excluded_fields,
                                fields_order=self.fields_order,
                            ),
                        ],
                    ),
                ],
            ),
            value="item",
            styles={
                "control": {"padding": "0.5rem"},
                "label": {"padding": 0},
                "item": {
                    "border": "1px solid color-mix(in srgb, var(--mantine-color-gray-light), transparent 40%)",
                    "background": "color-mix(in srgb, var(--mantine-color-gray-light), transparent 80%)",
                    "marginBottom": "0.5rem",
                    "borderRadius": "0.25rem",
                },
                "content": {
                    "display": "flex",
                    "flexDirection": "column",
                    "gap": "0.375rem",
                    "padding": "0.125rem 0.5rem 0.5rem",
                },
            },
            style={"--pydf-field-cols": "var(--pydf-form-cols)"},
            className="pydantic-form-field",
        )

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Create a form field of type checklist to interact with the model field."""
        try:
            renderer = getattr(self, f"{self.render_type}_item")
        except AttributeError as exc:
            raise ValueError(f"Unknown render type: {self.render_type}") from exc
        return renderer(item, aio_id, form_id, field, parent, field_info)

    # Open a model modal when editing an item
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="syncTrue"),
        Output(ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened", allow_duplicate=True),
        Input(ids.edit(MATCH, MATCH, MATCH, MATCH, MATCH), "n_clicks"),
        prevent_initial_call=True,
    )

    # Close a model modal when saving an item
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="syncFalse"),
        Output(ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened", allow_duplicate=True),
        Input(ids.modal_save(MATCH, MATCH, MATCH, MATCH, MATCH), "n_clicks"),
        prevent_initial_call=True,
    )



import gettext
import os
from contextlib import contextmanager
from functools import partial
from pathlib import Path

DEFAULT_LANGUAGE = "en"
SUPPORTED_LANGUAGES = ["en", "fr"]

locale_dir = str(Path(__file__).parent / "locales")
gettext.bindtextdomain("pydf", locale_dir)


@contextmanager
def language_context(language: str):
    """Language context for pydf."""
    if language:
        language_before = os.getenv("LANGUAGE")
        os.environ["LANGUAGE"] = language
        try:
            yield
        finally:
            if language_before is None:
                del os.environ["LANGUAGE"]
            else:
                os.environ["LANGUAGE"] = language_before
    else:
        yield


_ = partial(gettext.dgettext, "pydf")
